window.onerror = (msg, url, line, column, error) => {
  window.github._sendMessage('error', {
    message: msg,
    url: url,
    line: line,
    column: column,
    error: JSON.stringify(error)
  })
}

/**
 * @typedef {object} Rect
 * @property {number} X - The X coordinate
 * @property {number} Y - The Y coordinate
 * @property {number} Width - The width
 * @property {number} Height - The height
 */

/**
 * Object for configuring and controlling the page content
 */
class GitHub {
  /**
   * Initialize
   */
  constructor () {
  }

  /**
   * Send a message to the native app.
   *
   * @param {string} name Name of the message
   * @param {object} message Message content
   */
  _sendMessage (name, message) {
    const payload = message
    payload.messageName = name
    payload.id = this._id

    window.native.sendMessage(JSON.stringify(payload))
  }

  /**
   * Listen for events and send a height event when they fire
   */
  _addHeightListener () {
    if (typeof ResizeObserver !== 'undefined') {
      const resizeObserver = new ResizeObserver(entries => {
        const height = entries[0].target.clientHeight
        this._updateHeight(height)
      })
      resizeObserver.observe(document.body)
    } else {
      this._addElementHeightListener('details', 'toggle')
      this._addElementHeightListener('video', 'resize')
      this._addElementHeightListener('img, svg, video', 'load')
      this._addElementHeightListener('img, svg, video', 'error')
      this._updateHeight(document.body.scrollHeight)
    }
  }

  /**
   * If ResizeObserver is undefined, we have to observe the height of different elements.
   * This is **not** the preferred method as it does not account for all height changes by default.
   * Fixes: https://github.com/github/mobile-android/issues/3867
   *
   * @param {string} querySelector Query selector for elements
   * @param {string} eventName Event name to listen to
   */
  _addElementHeightListener (querySelector, eventName) {
    const elements = Array.from(document.querySelectorAll(querySelector))
    for (const element of elements) {
      element.addEventListener(eventName, () => {
        const height = document.body.scrollHeight
        this._updateHeight(height)
      })
    }
  }

  /**
   * Send a height to the app
   *
   * @param {int} height The new height to send to the app
   */
  _updateHeight (height) {
    if (height === this._lastHeight) {
      return
    }

    this._lastHeight = height
    this._sendMessage('height', { height: height })
  }

  /**
   * Unescape `?raw=true` at the end of image URLs
   */
  _unescapeImageURLs () {
    const images = Array.from(document.querySelectorAll('img'))
    for (const image of images) {
      const url = image.src
      if (!url || !url.endsWith('%3Fraw%3Dtrue')) {
        continue
      }

      image.src = url.replace(/%3Fraw%3Dtrue$/, '?raw=true')
    }
  }

  /**
   * Post-process and setup listeners after changing the main content
   */
  _didLoad () {
    // Clear the height cache
    this._lastHeight = 0

    // Post-process content
    this._unescapeImageURLs()

    // Add a height listener
    this._addHeightListener()
  }

  /**
   * Load content into the web view
   *
   * @param {string} id
   * @param {string} html Content to load
   * @param {boolean} isCommitSuggestedChangesEnabled to indicate if we show the commit button
   * @param {boolean} isTaskListCompletionEnabled to indicate if we should enable task list checkboxes
   * @param {overridePos} checkbox position to override the check state on
   * @param {overrideVal} checked state for {overridePos}
   */
  load (id, html, isCommitSuggestedChangesEnabled, isTaskListCompletionEnabled, overridePos, overrideVal) {
    this._id = id
    const content = document.getElementById('content')

    if (content.innerHTML === html && content.getAttribute('overridePos') === overridePos) {
      return
    }

    content.setAttribute('overridePos', overridePos)

    // Set the content
    content.innerHTML = html

    // Add a commit suggested changes button
    if (isCommitSuggestedChangesEnabled) {
      for (const suggestedChangeElement of document.getElementsByClassName('js-suggested-changes-blob')) {
        const buttonContainers = suggestedChangeElement.getElementsByClassName('js-apply-changes')
        if (buttonContainers && buttonContainers.length > 0) {
          this._addSuggestedChangesButton(
            buttonContainers[0],
            suggestedChangeElement.id,
            suggestedChangeElement.outerHTML
          )
        }
      }
    }

    const checkBoxes = document.getElementsByClassName('task-list-item-checkbox')
    for (let i = 0; i < checkBoxes.length; i++) {
        if(i == overridePos){
            checkBoxes[i].checked = overrideVal === 'true'
        }
        if(isTaskListCompletionEnabled){
            checkBoxes[i].disabled = false
            checkBoxes[i].addEventListener('change', (event) => {
                this._commitTaskItemChanged(i, event.target.checked)
                });
        }
        else{
            checkBoxes[i].disabled = true
        }
    }
    this._didLoad()
  }

  /**
   * Load a suggested changes button into the given element
   *
   * @param {element} container to add the suggested changes button
   * @param {id} the suggested change id
   * @param {previewHTML} the full suggested change HTML
   */
  _addSuggestedChangesButton (container, id, previewHTML) {
    container.innerHTML = "<div class=\"p-2 border-top\">" +
      "<button onclick=\"window.github._commitSuggestedChange('" + id + "', '" + escape(previewHTML) + "')\" " +
      "style=\"" +
      "padding: 10px 0px; " +
      "margin: 2px 0px; " +
      "display: flex; " +
      "align-items: center; " +
      "justify-content: center; " +
      "width: 100%; " +
      "text-transform: uppercase; " +
      "font-weight: 500; " +
      "font-size: 14px; " +
      "letter-spacing: 0.05em; " +
      "background-color: var(--backgroundElevatedSecondary); " +
      "color: var(--blue-500); " +
      "border-color: var(--border); " +
      "border-radius: 6px; " +
      "border-width: 1px; " +
      "outline: none; " +
      "\">" +
      "<svg style=\"fill: var(--blue-500); margin: 4px; 0px; 0px; 0px;\" " +
      "viewBox=\"0 0 16 16\" width=\"16\" height=\"16\"><path fill-rule=\"evenodd\" " +
      "d=\"M10.5 7.75a2.5 2.5 0 11-5 0 2.5 2.5 0 015 0zm1.43.75a4.002 4.002 0 01-7.86 0H.75a.75.75 0 " +
      "110-1.5h3.32a4.001 4.001 0 017.86 0h3.32a.75.75 0 110 1.5h-3.32z\"></path></svg>" +
      "Commit" +
      "</button>" +
      "</div>"
  }

  /**
   * Commit the suggested change back in the native view
   *
   * @param {id} the suggested change id
   * @param {escapedPreviewHTML} the full, escaped suggested change HTML
   */
  _commitSuggestedChange (id, escapedPreviewHTML) {
    this._sendMessage('commit_suggestion', {
      suggestionId: id,
      previewHTML: unescape(escapedPreviewHTML)
    })
  }

    /**
     * change the task list checkbox value
     *
     * @param {position} the position of the checkbox in the task list
     * @param {checked} checked value
     */
    _commitTaskItemChanged (position, checked) {
      this._sendMessage('task_changed', {
        taskPosition: position,
        taskChecked: checked
      })
    }

  /**
   * Load content into the web view
   *
   * @param {anchor} anchor
   */
  getAnchorPosition (anchor) {
    var position = this._positionOf(anchor)
    this._sendMessage('scroll_to', {
        posY: position,
        anchor: anchor
    })
  }

  /**
   * Get the position of an element by ID
   *
   * @param {string} elementID Name of the anchor tapped
   * @returns {number} The element's top Y position
   */
  _positionOf (elementID) {
    const element = this._getElementById('user-content-' + elementID)
    if (element === null) {
      return 0.0
    }
    const rect = element.getBoundingClientRect()

    return rect.top
  }

  /**
   * Get the DOM element by the given id in a case-insensitive way. We first try
   * to find an element which as the exact given id. If we cannot find any, we
   * search again with lowercased id.
   *
   * @param {string} id of the element that we want to find
   * @returns {HTMLElement} element or `null` if not found.
   */
  _getElementById (id) {
    const element = document.getElementById(id)
    if (element === null) {
      return document.getElementById(id.toLowerCase())
    }
    return element
  }
}

const start = () => {
  window.github = new GitHub()
}
if (document.readyState === 'loading') {
  document.addEventListener('DOMContentLoaded', start)
} else {
  start()
}

if (typeof exports !== 'undefined') {
  exports.GitHub = GitHub
}
